@@ -68,14 +68,45 @@ impl std::ops::BitAnd<Qop> for QopSet {
 /// Most of the information here is taken from the `WWW-Authenticate` or
 /// `Proxy-Authenticate` header. This also internally maintains a nonce counter.
 ///
-/// Implementation notes:
+/// ## Implementation notes
+///
+/// *   Recalculates `H(A1)` on each [`DigestClient::respond`] call. It'd be
+///     more CPU-efficient to calculate `H(A1)` only once by supplying the
+///     username and password at construction time or by caching (username,
+///     password) -> `H(A1)` mappings internally. `DigestClient` prioritizes
+///     simplicity instead.
 /// *   Always responds using `UTF-8`, and thus doesn't use or keep around the `charset`
 ///     parameter. The RFC only allows that parameter to be set to `UTF-8` anyway.
-/// *   Doesn't support `userhash`. This option is simple enough for a client to implement,
-///     but I'd be surprised to see a server that uses it. It appears to require the server
-///     to perform `O(users)` cryptographic operations to find the correct user to operate on.
-///     Needless to say, that's unrealistic with a large authentication database.
-/// *   Supports RFC 2069 compatibility, even though RFC 7616 drops it.
+/// *   Supports [RFC 2069](https://datatracker.ietf.org/doc/html/rfc2069) compatibility as in
+///     [RFC 2617 section 3.2.2.1](https://datatracker.ietf.org/doc/html/rfc2617#section-3.2.2.1),
+///     even though RFC 7616 drops it. There are still RTSP cameras being sold
+///     in 2021 that use the RFC 2069-style calculations.
+/// *   Supports RFC 7616 `userhash`, even though it seems impractical and only
+///     marginally useful. The server must index the userhash for each supported
+///     algorithm or calculate it on-the-fly for all users in the database.
+/// *   The `-sess` algorithm variants haven't been tested; there's no example
+///     in the RFCs.
+///
+/// ## Security considerations
+///
+/// We strongly advise *servers* against implementing `Digest`:
+///
+/// *   It's actively harmful in that it prevents the server from securing their
+///     password storage via salted password hashes. See [RFC 7616 Section
+///     5.2](https://datatracker.ietf.org/doc/html/rfc7616#section-5.2).
+///     When your server offers `Digest` authentication, it is advertising that
+///     it stores plaintext passwords!
+/// *   It's no replacement for TLS in terms of protecting confidentiality of
+///     the password, much less confidentiality of any other information.
+///
+/// For *clients*, when a server supports both `Digest` and `Basic`, we advise
+/// using `Digest`. It provides (slightly) more confidentiality of passwords
+/// over the wire.
+///
+/// Some servers *only* support `Digest`. E.g.,
+/// [ONVIF](https://www.onvif.org/profiles/specifications/) mandates the
+/// `Digest` scheme. It doesn't prohibit implementing other schemes, but some
+/// cameras meet the specification's requirement and do no more.
 #[derive(Eq, PartialEq)]
 pub struct DigestClient {
     /// Holds unescaped versions of all string fields.
@@ -88,21 +119,21 @@ pub struct DigestClient {
     /// 2.  `domain`: `[domain_start, opaque_start)`
     /// 3.  `opaque`: `[opaque_start, nonce_start)`
     /// 4.  `nonce`: `[nonce_start, buf.len())`
-    ///
-    /// `nonce` is last so that future support for parsing the `Authentication-Info` response's
-    /// `next_nonce` field can easily rewrite `nonce` without modifying the other fields.
-    buf: String,
+    buf: Box<str>,
 
-    // Positions described in `buf` comment above. See respective methods' doc comments
-    // for more information.
-    domain_start: u32,
-    opaque_start: u32,
-    nonce_start: u32,
+    // Positions described in `buf` comment above. See respective methods' doc
+    // comments for more information. These are stored as `u16` to save space,
+    // and because it's unreasonable for them to be large.
+    domain_start: u16,
+    opaque_start: u16,
+    nonce_start: u16,
 
     // Non-string fields. See respective methods' doc comments for more information.
     algorithm: Algorithm,
+    session: bool,
     stale: bool,
     rfc2069_compat: bool,
+    userhash: bool,
     qop: QopSet,
     nc: u32,
 }
@@ -160,8 +191,11 @@ impl DigestClient {
         self.stale
     }
 
-    /// Returns true if using RFC 2069 compatibility mode, in which the
-    /// `request-digest` is calculated without the nonce count, conce, or qop.
+    /// Returns true if using [RFC 2069](https://datatracker.ietf.org/doc/html/rfc2069)
+    /// compatibility mode as in [RFC 2617 section
+    /// 3.2.2.1](https://datatracker.ietf.org/doc/html/rfc2617#section-3.2.2.1).
+    ///
+    /// If so, `request-digest` is calculated without the nonce count, conce, or qop.
     #[inline]
     pub fn rfc2069_compat(&self) -> bool {
         self.rfc2069_compat
@@ -173,6 +207,12 @@ impl DigestClient {
         self.algorithm
     }
 
+    /// Returns if the session style `A1` will be used.
+    #[inline]
+    pub fn session(&self) -> bool {
+        self.session
+    }
+
     /// Returns the acceptable `qop` (quality of protection) values.
     #[inline]
     pub fn qop(&self) -> QopSet {
@@ -216,13 +256,22 @@ impl DigestClient {
     /// and have `respond` delegate to the testing version. We don't do that because
     fn respond_inner(&mut self, p: &PasswordParams, cnonce: &str) -> Result<String, String> {
         let realm = self.realm();
-        let h_a1 = self.algorithm.h(&[
+        let mut h_a1 = self.algorithm.h(&[
             p.username.as_bytes(),
             b":",
             realm.as_bytes(),
             b":",
             p.password.as_bytes(),
         ]);
+        if self.session {
+            h_a1 = self.algorithm.h(&[
+                h_a1.as_bytes(),
+                b":",
+                self.nonce().as_bytes(),
+                b":",
+                cnonce.as_bytes(),
+            ]);
+        }
 
         // Select the best available qop and calculate H(A2) as in
         // [https://datatracker.ietf.org/doc/html/rfc7616#section-3.4.3].
@@ -241,9 +290,9 @@ impl DigestClient {
             return Err("no supported/available qop".into());
         }
 
-        self.nc = self.nc.checked_add(1).ok_or("nonce count exhausted")?;
+        let nc = self.nc.checked_add(1).ok_or("nonce count exhausted")?;
         let mut hex_nc = [0u8; 8];
-        let _ = write!(&mut hex_nc[..], "{:08x}", self.nc);
+        let _ = write!(&mut hex_nc[..], "{:08x}", nc);
         let str_hex_nc = match std::str::from_utf8(&hex_nc[..]) {
             Ok(h) => h,
             Err(_) => unreachable!(),
@@ -276,7 +325,13 @@ impl DigestClient {
 
         let mut out = String::with_capacity(128);
         out.push_str("Digest ");
-        if is_valid_quoted_value(p.username) {
+        if self.userhash {
+            let hashed = self
+                .algorithm
+                .h(&[p.username.as_bytes(), b":", realm.as_bytes()]);
+            append_quoted_key_value(&mut out, "username", &hashed)?;
+            append_unquoted_key_value(&mut out, "userhash", "true");
+        } else if is_valid_quoted_value(p.username) {
             append_quoted_key_value(&mut out, "username", p.username)?;
         } else {
             append_extended_key_value(&mut out, "username", p.username);
@@ -285,7 +340,7 @@ impl DigestClient {
         append_quoted_key_value(&mut out, "uri", p.uri)?;
         append_quoted_key_value(&mut out, "nonce", self.nonce())?;
         if !self.rfc2069_compat {
-            append_unquoted_key_value(&mut out, "algorithm", self.algorithm.as_str());
+            append_unquoted_key_value(&mut out, "algorithm", self.algorithm.as_str(self.session));
             append_unquoted_key_value(&mut out, "nc", str_hex_nc);
             append_quoted_key_value(&mut out, "cnonce", cnonce)?;
             append_unquoted_key_value(&mut out, "qop", qop.as_str());
@@ -295,6 +350,7 @@ impl DigestClient {
             append_quoted_key_value(&mut out, "opaque", o)?;
         }
         out.truncate(out.len() - 2); // remove final ", "
+        self.nc = nc;
         Ok(out)
     }
 }
@@ -310,13 +366,15 @@ impl TryFrom<&ChallengeRef<'_>> for DigestClient {
             ));
         }
         let mut buf_len = 0;
+        let mut unused_len = 0;
         let mut realm = None;
         let mut domain = None;
         let mut nonce = None;
         let mut opaque = None;
         let mut stale = false;
-        let mut algorithm = None;
+        let mut algorithm_and_session = None;
         let mut qop_str = None;
+        let mut userhash_str = None;
 
         // Parse response header field parameters as in
         // [https://datatracker.ietf.org/doc/html/rfc7616#section-3.3].
@@ -329,44 +387,32 @@ impl TryFrom<&ChallengeRef<'_>> for DigestClient {
                 || store_param(k, v, "domain", &mut domain, &mut buf_len)?
                 || store_param(k, v, "nonce", &mut nonce, &mut buf_len)?
                 || store_param(k, v, "opaque", &mut opaque, &mut buf_len)?
-                || store_param(k, v, "qop", &mut qop_str, &mut buf_len)?
+                || store_param(k, v, "qop", &mut qop_str, &mut unused_len)?
+                || store_param(k, v, "userhash", &mut userhash_str, &mut unused_len)?
             {
                 // Do nothing here.
             } else if k.eq_ignore_ascii_case("stale") {
-                stale = v.raw.eq_ignore_ascii_case("true");
+                stale = v.escaped.eq_ignore_ascii_case("true");
             } else if k.eq_ignore_ascii_case("algorithm") {
-                algorithm = Some(Algorithm::parse(v.raw)?);
+                algorithm_and_session = Some(Algorithm::parse(v.escaped)?);
             }
         }
         let realm = realm.ok_or("missing required parameter realm")?;
         let nonce = nonce.ok_or("missing required parameter nonce")?;
-        if buf_len > u32::MAX as usize {
+        if buf_len > u16::MAX as usize {
             // Incredibly unlikely, but just for completeness.
             return Err(format!(
-                "Unescaped parameters' length {} exceeds u32::MAX!",
+                "Unescaped parameters' length {} exceeds u16::MAX!",
                 buf_len
             ));
         }
 
-        let algorithm = algorithm.unwrap_or(Algorithm::Md5);
+        let algorithm_and_session = algorithm_and_session.unwrap_or((Algorithm::Md5, false));
 
         let mut buf = String::with_capacity(buf_len);
-        realm.append_unescaped(&mut buf);
-        let domain_start = buf.len();
-        if let Some(d) = domain {
-            d.append_unescaped(&mut buf);
-        }
-        let opaque_start = buf.len();
-        if let Some(o) = opaque {
-            o.append_unescaped(&mut buf);
-        }
-        let nonce_start = buf.len();
-        nonce.append_unescaped(&mut buf);
         let mut qop = QopSet(0);
         let rfc2069_compat = if let Some(qop_str) = qop_str {
-            let qop_start = buf.len();
-            qop_str.append_unescaped(&mut buf);
-            let qop_str = &buf[qop_start..];
+            let qop_str = qop_str.unescaped_with_scratch(&mut buf);
             for v in qop_str.split(',') {
                 let v = v.trim();
                 if v.eq_ignore_ascii_case("auth") {
@@ -378,7 +424,7 @@ impl TryFrom<&ChallengeRef<'_>> for DigestClient {
             if qop.0 == 0 {
                 return Err(format!("no supported qop in {:?}", qop_str));
             }
-            buf.truncate(qop_start);
+            buf.clear();
             false
         } else {
             // An absent qop is treated as "auth", according to
@@ -386,15 +432,36 @@ impl TryFrom<&ChallengeRef<'_>> for DigestClient {
             qop.0 |= Qop::Auth as u8;
             true
         };
+        let userhash;
+        if let Some(userhash_str) = userhash_str {
+            let userhash_str = userhash_str.unescaped_with_scratch(&mut buf);
+            userhash = userhash_str.eq_ignore_ascii_case("true");
+            buf.clear();
+        } else {
+            userhash = false;
+        };
+        realm.append_unescaped(&mut buf);
+        let domain_start = buf.len();
+        if let Some(d) = domain {
+            d.append_unescaped(&mut buf);
+        }
+        let opaque_start = buf.len();
+        if let Some(o) = opaque {
+            o.append_unescaped(&mut buf);
+        }
+        let nonce_start = buf.len();
+        nonce.append_unescaped(&mut buf);
         Ok(DigestClient {
-            buf,
-            domain_start: domain_start as u32,
-            opaque_start: opaque_start as u32,
-            nonce_start: nonce_start as u32,
-            algorithm,
+            buf: buf.into_boxed_str(),
+            domain_start: domain_start as u16,
+            opaque_start: opaque_start as u16,
+            nonce_start: nonce_start as u16,
+            algorithm: algorithm_and_session.0,
+            session: algorithm_and_session.1,
             stale,
-            qop,
             rfc2069_compat,
+            userhash,
+            qop,
             nc: 0,
         })
     }
@@ -407,10 +474,11 @@ impl std::fmt::Debug for DigestClient {
             .field("domain", &self.domain())
             .field("opaque", &self.opaque())
             .field("nonce", &self.nonce())
-            .field("algorithm", &self.algorithm)
+            .field("algorithm", &self.algorithm.as_str(self.session))
             .field("stale", &self.stale)
             .field("qop", &self.qop)
             .field("rfc2069_compat", &self.rfc2069_compat)
+            .field("userhash", &self.userhash)
             .field("nc", &self.nc)
             .finish()
     }
@@ -491,6 +559,9 @@ fn append_quoted_key_value(out: &mut String, key: &str, value: &str) -> Result<(
 
 /// Supported algorithm from the [HTTP Digest Algorithm Values
 /// registry](https://www.iana.org/assignments/http-dig-alg/http-dig-alg.xhtml).
+///
+/// This doesn't store whether the session variant (`<Algorithm>-sess`) was
+/// requested; see [`DigestClient::session`] for that.
 #[derive(Copy, Clone, Debug, Eq, PartialEq)]
 #[non_exhaustive]
 pub enum Algorithm {
@@ -500,23 +571,28 @@ pub enum Algorithm {
 }
 
 impl Algorithm {
-    fn parse(s: &str) -> Result<Self, String> {
-        if s.to_ascii_lowercase() == "md5" {
-            Ok(Algorithm::Md5)
-        } else if s.to_ascii_lowercase() == "sha-256" {
-            Ok(Algorithm::Sha256)
-        } else if s.to_ascii_lowercase() == "sha-512-256" {
-            Ok(Algorithm::Sha512Trunc256)
-        } else {
-            Err(format!("unknown algorithm {:?}", s))
-        }
+    /// Parses a string into a tuple of `Algorithm` and a bool representing
+    /// whether the `-sess` suffix is present.
+    fn parse(s: &str) -> Result<(Self, bool), String> {
+        Ok(match s {
+            "MD5" => (Algorithm::Md5, false),
+            "MD5-sess" => (Algorithm::Md5, true),
+            "SHA-256" => (Algorithm::Sha256, false),
+            "SHA-256-sess" => (Algorithm::Sha256, true),
+            "SHA-512-256" => (Algorithm::Sha512Trunc256, false),
+            "SHA-512-256-sess" => (Algorithm::Sha512Trunc256, true),
+            _ => return Err(format!("unknown algorithm {:?}", s)),
+        })
     }
 
-    fn as_str(&self) -> &'static str {
-        match self {
-            Algorithm::Md5 => "MD5",
-            Algorithm::Sha256 => "SHA-256",
-            Algorithm::Sha512Trunc256 => "SHA-512-256",
+    fn as_str(&self, session: bool) -> &'static str {
+        match (self, session) {
+            (Algorithm::Md5, false) => "MD5",
+            (Algorithm::Md5, true) => "MD5-sess",
+            (Algorithm::Sha256, false) => "SHA-256",
+            (Algorithm::Sha256, true) => "SHA-256-sess",
+            (Algorithm::Sha512Trunc256, false) => "SHA-512-256",
+            (Algorithm::Sha512Trunc256, true) => "SHA-512-256-sess",
         }
     }
 
@@ -625,6 +701,66 @@ mod tests {
                     response=\"8ca523f5e9506fed4657c9700eebdbec\", \
                     opaque=\"FQhe/qaU925kfnzjCev0ciny7QMkPqMAFRtzCUYo5tdS\""
         );
+        assert_eq!(ctxs[1].nc, 1);
+    }
+
+    /// Tests a made-up example with `MD5-sess`. There's no example in the RFC,
+    /// and these values haven't been tested against any other implementation.
+    /// But having the test here ensures we don't accidentally change the
+    /// algorithm.
+    #[test]
+    fn md5_sess() {
+        let www_authenticate = "\
+            Digest \
+            realm=\"http-auth@example.org\", \
+            qop=\"auth, auth-int\", \
+            algorithm=MD5-sess, \
+            nonce=\"7ypf/xlj9XXwfDPEoM4URrv/xwf94BcCAzFZH4GiTo0v\", \
+            opaque=\"FQhe/qaU925kfnzjCev0ciny7QMkPqMAFRtzCUYo5tdS\"";
+        let challenges = dbg!(crate::parse_challenges(www_authenticate).unwrap());
+        assert_eq!(challenges.len(), 1);
+        let ctxs: Result<Vec<_>, _> = challenges.iter().map(DigestClient::try_from).collect();
+        let mut ctxs = dbg!(ctxs.unwrap());
+        assert_eq!(ctxs[0].realm(), "http-auth@example.org");
+        assert_eq!(ctxs[0].domain(), "");
+        assert_eq!(
+            ctxs[0].nonce(),
+            "7ypf/xlj9XXwfDPEoM4URrv/xwf94BcCAzFZH4GiTo0v"
+        );
+        assert_eq!(
+            ctxs[0].opaque(),
+            Some("FQhe/qaU925kfnzjCev0ciny7QMkPqMAFRtzCUYo5tdS")
+        );
+        assert_eq!(ctxs[0].stale(), false);
+        assert_eq!(ctxs[0].algorithm(), Algorithm::Md5);
+        assert_eq!(ctxs[0].session(), true);
+        assert_eq!(ctxs[0].qop().0, (Qop::Auth as u8) | (Qop::AuthInt as u8));
+        assert_eq!(ctxs[0].nonce_count(), 0);
+        let params = crate::PasswordParams {
+            username: "Mufasa",
+            password: "Circle of Life",
+            uri: "/dir/index.html",
+            body: None,
+            method: "GET",
+        };
+        assert_eq!(
+            &mut ctxs[0]
+                .respond_with_testing_cnonce(
+                    &params,
+                    "f2/wE4q74E6zIJEtWaHKaf5wv/H5QzzpXusqGemxURZJ"
+                )
+                .unwrap(),
+            "Digest username=\"Mufasa\", \
+                    realm=\"http-auth@example.org\", \
+                    uri=\"/dir/index.html\", \
+                    nonce=\"7ypf/xlj9XXwfDPEoM4URrv/xwf94BcCAzFZH4GiTo0v\", \
+                    algorithm=MD5-sess, \
+                    nc=00000001, \
+                    cnonce=\"f2/wE4q74E6zIJEtWaHKaf5wv/H5QzzpXusqGemxURZJ\", \
+                    qop=auth, \
+                    response=\"e783283f46242139c486a698fec7211d\", \
+                    opaque=\"FQhe/qaU925kfnzjCev0ciny7QMkPqMAFRtzCUYo5tdS\""
+        );
         assert_eq!(ctxs[0].nc, 1);
     }
 
@@ -657,6 +793,7 @@ mod tests {
             Some("HRPCssKJSGjCrkzDg8OhwpzCiGPChXYjwrI2QmXDnsOS")
         );
         assert_eq!(ctxs[0].stale, false);
+        assert_eq!(ctxs[0].userhash, true);
         assert_eq!(ctxs[0].algorithm, Algorithm::Sha512Trunc256);
         assert_eq!(ctxs[0].qop.0, Qop::Auth as u8);
         assert_eq!(ctxs[0].nc, 0);
@@ -670,6 +807,30 @@ mod tests {
 
         // Note the username and response values in the RFC are *wrong*!
         // https://www.rfc-editor.org/errata/eid4897
+        assert_eq!(
+            &mut ctxs[0]
+                .respond_with_testing_cnonce(
+                    &params,
+                    "NTg6RKcb9boFIAS3KrFK9BGeh+iDa/sm6jUMp2wds69v"
+                )
+                .unwrap(),
+            "\
+            Digest \
+            username=\"793263caabb707a56211940d90411ea4a575adeccb7e360aeb624ed06ece9b0b\", \
+            userhash=true, \
+            realm=\"api@example.org\", \
+            uri=\"/doe.json\", \
+            nonce=\"5TsQWLVdgBdmrQ0XsxbDODV+57QdFR34I9HAbC/RVvkK\", \
+            algorithm=SHA-512-256, \
+            nc=00000001, \
+            cnonce=\"NTg6RKcb9boFIAS3KrFK9BGeh+iDa/sm6jUMp2wds69v\", \
+            qop=auth, \
+            response=\"3798d4131c277846293534c3edc11bd8a5e4cdcbff78b05db9d95eeb1cec68a5\", \
+            opaque=\"HRPCssKJSGjCrkzDg8OhwpzCiGPChXYjwrI2QmXDnsOS\""
+        );
+        assert_eq!(ctxs[0].nc, 1);
+        ctxs[0].userhash = false;
+        ctxs[0].nc = 0;
         assert_eq!(
             &mut ctxs[0]
                 .respond_with_testing_cnonce(
